// CodeView.cpp : implementation file
//

#include "stdafx.h"
#include "HardCoreWookie.h"
#include "HardCoreWookieDoc.h"
#include "CodeView.h"
#include "../emu/hc11.h"
#include <string.h>
#include <ctype.h>
#include <fstream.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CCodeView

IMPLEMENT_DYNCREATE(CCodeView, CEditView)

CCodeView::CCodeView()
{
}

CCodeView::~CCodeView()
{
}


BEGIN_MESSAGE_MAP(CCodeView, CEditView)
	//{{AFX_MSG_MAP(CCodeView)
	ON_WM_CREATE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CCodeView drawing

void CCodeView::OnDraw(CDC* pDC)
{
	CDocument* pDoc = GetDocument();
	// TODO: add draw code here
}

/////////////////////////////////////////////////////////////////////////////
// CCodeView diagnostics

#ifdef _DEBUG
void CCodeView::AssertValid() const
{
	CEditView::AssertValid();
}

void CCodeView::Dump(CDumpContext& dc) const
{
	CEditView::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CCodeView message handlers

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: OnCreate
//
// Author: Matt Enwald
//
// Purpose: To prepare the Code View when a new lst/s19 file is opened.
//
// Input Parameters: None
//
// Return Value: None
//
// Preconditions: None
//
// Post Conditions: The copy of the lst file in the code view must be read-
//				    only so the user can't delete or modify the code.  
//					The CCodeView class must be notified that the lst/s19
//					file has been opened.
//
/////////////////////////////////////////////////////////////////////////////
int CCodeView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CEditView::OnCreate(lpCreateStruct) == -1)
		return -1;
	
//	Make window read-only

	// Set is file open to closed
	m_isFileOpen = false;

	//Save this pointer in Doc
	((CHardCoreWookieDoc *)GetDocument())->m_pCodeView=this;	

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: LoadlstFile
//
// Author: Matt Enwald
//
// Purpose: To load in the s19/lst file that the user specifies from the open
//			common dialog box.
//
// Input Parameters: FILE* lstfile - a pointer to the open lst file that the 
//					 user wants in the code view.
//
// Return Value: None
//
// Preconditions: The input file must be open and have valid data
//
// Post Conditions: None
//
/////////////////////////////////////////////////////////////////////////////
bool CCodeView::LoadlstFile(CString lstpath)
{
	CString bigbuffer;
	char littlebuffer[101];
	int len;
    bool status = false;
	
    FILE* lstfile = fopen(lstpath,"rb");
	
	GetEditCtrl().SetReadOnly(false);

	if (lstfile != NULL)
    {		 	
		GetEditCtrl().SetWindowText(NULL);
		// Read in from the lst file passed in as a parameter
	    while (!feof(lstfile))
	    {
		    len = fread(littlebuffer,1,sizeof(littlebuffer)-1,lstfile);		
		    littlebuffer[len] = 0;
		    bigbuffer += littlebuffer;
	    }
        
        // GCC listing files are different.
        if(bigbuffer.Find("Start MC6811 gcc assembly output")!=-1)
        {
            lst_mode = GCC_LST;
        }
        else lst_mode = AS11M_LST;
        
		GetEditCtrl().FmtLines(true);
        // Display lst file in the code view window
	    SetWindowText(bigbuffer);

	    // Set is file open to open.  This flag is used to enable and disable features
	    // like the run and step buttons.
	    m_isFileOpen = true;

        fclose(lstfile);	
        status = true;
    }
	else
	{
		AfxMessageBox("Could not open list file.\r\nPlease verify that the list file is in the\n\rsame directory as the .s19 file.");		        
	}		

	GetEditCtrl().SetReadOnly(true);
    return status;
}

bool CCodeView::LoadasmFile(CString path)
{
	bool status = false;
	CString bigbuffer;
	char littlebuffer[101];
	int len;
    
	asmpath = path;
	
    FILE* asmfile = fopen(asmpath,"rb");
	if (asmfile != NULL)
    {		 	
		GetEditCtrl().SetWindowText(NULL);
        // Read in from the lst file passed in as a parameter
	    while (!feof(asmfile))
	    {
		    len = fread(littlebuffer,1,sizeof(littlebuffer)-1,asmfile);		
		    littlebuffer[len] = 0;
		    bigbuffer += littlebuffer;
	    }
        fclose(asmfile);

        // Display lst file in the code view window
	    SetWindowText(bigbuffer);
	    m_isFileOpen = true;        	
        status = true;
		GetEditCtrl().SetReadOnly(false);
    }
	else
	{
		AfxMessageBox("Could not open asm file.\r\n");		        
	}		
 
   return status;
}


bool CCodeView::SaveasmFile(void)
{   
	ofstream outf(asmpath);
	CString bigbuffer;

	if(!outf)
		return false;

	GetWindowText(bigbuffer);
	
	outf << bigbuffer;

	return true;
/*	
	int len,i,wlen;
  
	char ch;
    FILE* asmfile = fopen(asmpath,"wb");
    const char *buffer; 
	
	if (asmfile != NULL)
    {			
		GetWindowText(bigbuffer);				
		
		len = bigbuffer.GetLength();
		buffer=bigbuffer.LockBuffer();					
		
		status = true;

		while(len)
		{
			wlen=fwrite(buffer,1,(len<1024)?len:1024,asmfile);		
			if(wlen == 0)
			{
				status = false;
				break;
			}
			else len -=wlen;
		}

		bigbuffer.UnlockBuffer();						
		fclose(asmfile);
		
		AfxMessageBox((status)?"Saved OK":"Save failed");
    }
	else
	{
		AfxMessageBox("Could not open file for output\n");
	}		
    return status;
	*/
}

void CCodeView::UnLoadFile(void)
{       				
        m_isFileOpen = false;			
		GetEditCtrl().SetWindowText(NULL);
		GetEditCtrl().SetReadOnly(true);		
}
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: SelectLine
//
// Authors: Matt Enwald, Kalle Anderson, Paul Clarke
//
// Purpose: To select the current line of code in the lst file according to 
//			the position of the program counter in the HC11
//
// Note: This works best on files assembled by AS11M. It is kludged to work
//       kinda on listing files generated by gcc. The program can only
//       consist of one file, and the starting address must have been
//       entered correctly. This would be a good place for futre improvement.
//
//
// Input Parameters: HC11* hc11 - a pointer to the core HC11
//
// Return Value: None
//
// Preconditions: The HC11 must be valid and in a valid state.
//
// Post Conditions: None
//
/////////////////////////////////////////////////////////////////////////////
void CCodeView::SelectLine(HC11 *hc11)
{
    const char *buffer = LockBuffer(); // LockBuffer returns a pointer to the 
 									   // CEditView's buffer (the lst file)
    const char *ptr;    
    unsigned int start=0,end=0,buflen = GetBufferLength();
	// GetBufferLength() returns the length of the string in the CEditView's buffer
    
	// Read the hex value of the HC11's program counter into hexstr so that we
	// know which line to select in the code view
    CString hexstr;    
    
    // quick fix for single lst file c programs made with GCC
    if(lst_mode == GCC_LST)
        hexstr.Format("%04X %02X",(int)(hc11->PC-start_addr), 
                                  (int)(hc11->memory.Read(hc11->PC)));	      
    else // AS11M lst file
        hexstr.Format("%04x %02x",(int)hc11->PC, (int)hc11->memory.Read(hc11->PC));	      

    
	// If the value of the PC is found in the lst file
    if((ptr = strstr(buffer, hexstr))!=NULL)
    {        
		// Set the highlight section from the beginning of the line containing the
		// PC to the end of the line, or to the beginning of a comment, excluding
		// white spaces
        end = start = ptr - buffer; 
//        while((end<buflen-1)&&(buffer[end+1]!='\n')&&(buffer[end+1]!=';'))end++;
//        while(isspace(buffer[end]))end--;
        while((start>0)&&(buffer[start-1]!='\n'))start--;
        while((end<buflen-1)&&(buffer[end+1]!='\n'))end++;
        UnlockBuffer();
        GetEditCtrl().SetSel(start,end+1,false);
    }
    // If the value of the PC is not found in the lst file
	else
    {
		UnlockBuffer();
        GetEditCtrl().SetSel(-1,true);        
        //hexstr.Format("PC points outside loaded program! ($%04x)",(int)hc11->PC);	      
        //AfxMessageBox(hexstr);                
    }    
}



//////////////////////////////////////////////////////////////////////////////
//
// Function Name: FindOrd
//
// Authors: Kalle Anderson
//
// Purpose: When the program is loaded up, the user must enter the starting
//          address. This will try to find the first org statement. The doc,
//          will stick this value in as the default in the startup dialog.
//
// Input Parameters: None
//
// Return Value: None
//
// Preconditions: The HC11 must be valid and in a valid state.
//
// Post Conditions: None
//
/////////////////////////////////////////////////////////////////////////////

word_t CCodeView::FindOrg(void)
{
    const char *buffer = LockBuffer(); // LockBuffer returns a pointer to the 
									  // CEditView's buffer (the lst file)
    const char *ptr;    
    word_t orgaddr = 0;
    
    if(((ptr=strstr(buffer, "org"))!=NULL) || 
	   ((ptr=strstr(buffer, "ORG"))!=NULL))
    {        
        ptr += 3; // o r g        
        while(isspace(*ptr))ptr++;   
        
        if(*ptr == '$') //hex
            sscanf(ptr+1,"%X",&orgaddr);
        else            //decimal
            sscanf(ptr,"%d",&orgaddr);       
    }    
    UnlockBuffer();
    return(orgaddr);
}


//////////////////////////////////////////////////////////////////////////////
//
// Function Name: AddRmbWatches
//
// Authors: Kalle Anderson
//
// Purpose: This fills the memory watch with all the rmb's in a program.
//          The rmb must be under 2 bytes. 
//          
// Input Parameters: The MemoryWatch window
//
// Return Value: None
//
/////////////////////////////////////////////////////////////////////////////

void CCodeView::AddRmbWatches(CMemoryWatch *watchwin)
{
    
    const char *buffer = LockBuffer(); // LockBuffer returns a pointer 
    const char *ptr=buffer;           //Man this code will be nasty :)
    const char *addr;
    const char *temp;

    word_t watch_address;    
    const int watch_name_length=20;      // personally I would have done                                        
    char watch_name[watch_name_length];  // char watch_name[20]
    int i =0, num_bytes;
            
    unsigned int buflen = GetBufferLength();

    // I have to apologize for this code, its probally kinda evil in a way
    while(((temp=strstr(ptr,"rmb"))!=NULL)||((temp=strstr(ptr,"fcb"))!=NULL))
    {        
        addr = temp;
        ptr=temp+3;
        if((unsigned)(ptr-buffer)>=buflen)break;
        if(*(addr-1)==' ')
        {                                
            if((sscanf(ptr,"%d",&num_bytes)!=0)&&(num_bytes<=2))
            {                
                while((*addr!='\n')&&(addr!=buffer))addr--;
                addr += 6; // should point at the address

                if(sscanf(addr,"%x",&watch_address)!=0)
                {           
                    while((!isspace(*addr))&&(addr!=(buffer+buflen)))addr++;
                    while(isspace(*addr)&&(addr!=(buffer+buflen)))addr++;
            
                    i=0;
                    while((!isspace(addr[i]))&&(i<10))
                    {
                        watch_name[i]=addr[i];
                        i++;
                    }   
                    
                    watch_name[i] = 0;
                    watchwin->AddWatch(watch_name,watch_address);

                    if(num_bytes==2)
                        watchwin->AddWatch(watch_name,watch_address+1);
                }
            }
        }
    }
    UnlockBuffer();        
}
