#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <io.h>

#define FALSE 0
#define TRUE 1

#include "as.h"
#include "table11.h"

char mapdn();
char *alloc();

static FILE *lstfile;

#define PAGE1   0x00

#define PAGE2   0x18

#define PAGE3   0x1A

#define PAGE4   0xCD

/* addressing modes */
#define IMMED   0
#define INDX    1
#define INDY    2
#define LIMMED  3       /* long immediate */
#define OTHER   4
int     yflag = 0;      /* YNOIMM, YLIMM, and CPD flag */
int fatalerr=0;


/*
 *      as ---  cross assembler main program
 */
int as11m_main(int argc,char **argv, FILE *of, FILE *lstf)
{
        char    **np;
        int     j = 0;

		fatalerr=0;

        Argv = argv;
        initialize();
        N_files = argc;
		Objfil = of;
		lstfile = lstf;		

        Lflag = 1;
        Cflag = 0;
        Sflag = 0;
        CREflag = 0;            

		root = NULL;

        Cfn = 0;
        np = argv;
        Line_num = 0; /* reset line number */
        while( ++Cfn <= N_files )
                if((Fd = fopen(*np++,"r")) == NULL)
                        fprintf(lstfile, "as: can't open %s\n",*np);
                else{
                        make_pass();
                        fclose(Fd);
                }
        if( Err_count == 0 ){
                Pass++;
                re_init();
                Cfn = 0;
                np = argv;
                Line_num = 0;
                while( ++Cfn <= N_files)
                        if((Fd = fopen(*np++,"r")) != NULL)
                            {
                                make_pass();
                                fclose(Fd);
                             }

                        if (Sflag == 1)
                          {
                            printf ("\f");
                            stable (root);
                          }
                        if (CREflag == 1)
                          {
                            printf ("\f");
                            cross (root);
                          }

						fprintf(Objfil,"S9030000FC\n"); /* at least give a decent ending */
                }        
		fprintf(lstfile, "%d errors\n", Err_count);
		return Err_count;
}
char	*hexstr = { "0123456789ABCDEF" } ;
/*
 *	any --- does str contain c?
 */

int any(char c,char *str)
{
	while(*str != EOS)
		if(*str++ == c)
			return(YES);
	return(NO);

}



/*
 *	mapdn --- convert A-Z to a-z
 */

char mapdn(char c)
{
	if( c >= 'A' && c <= 'Z')
		return(c+040);
	return(c);
}



/*
 *	lobyte --- return low byte of an int
 */

int lobyte(int i)
{
	return(i&0xFF);
}

/*
 *	hibyte --- return high byte of an int
 */

int hibyte(int i)
{
	return((i>>8)&0xFF);
}



/*

 *	head --- is str2 the head of str1?

 */

int head(char *str1,char *str2)
{

	while( *str1 != EOS && *str2 != EOS){

		if( *str1 != *str2 )break;

		str1++;

		str2++;

		}

	if(*str1 == *str2)return(YES);

	if(*str2==EOS)

		if( any(*str1," \t\n,+-];*") )return(YES);

	return(NO);

}



/*
 *	alpha --- is character a legal letter
 */

int alpha(char c)
{

	if( c<= 'z' && c>= 'a' )return(YES);

	if( c<= 'Z' && c>= 'A' )return(YES);

	if( c== '_' )return(YES);

	if( c== '.' )return(YES);

	return(NO);

}

/*
 *	alphan --- is character a legal letter or digit
 */

int alphan(char c)
{
	if( alpha(c) )return(YES);

	if( c<= '9' && c>= '0' )return(YES);

	if( c == '$' )return(YES);      /* allow imbedded $ */

	return(NO);
}



/*
 *	white --- is character whitespace?
 */
int white(char c)
{
	if( c == TAB || c == BLANK || c == '\n' )return(YES);

	return(NO);
}



/*
 *	alloc --- allocate memory
 */

char *alloc(int nbytes)
{
	return((char *)malloc(nbytes));
}

/*

 *      file I/O version of forward ref handler

 */

/*
 *	delim --- check if character is a delimiter
 */

int delim(char c)
{

	if( any(c," \t\n"))

		return(YES);

	return(NO);

}



/*
 *	skip_white --- move pointer to next non-whitespace char
 */
char *skip_white(char *ptr)
{
	while(*ptr==BLANK || *ptr==TAB)
		ptr++;
	return(ptr);
}

void hexout(int byte)
{
	byte = lobyte(byte);
	fprintf(Objfil,"%c%c",hexstr[byte>>4],hexstr[byte&017]);
}



/*
 *	f_record --- flush record out in `S1' format
 */
void f_record(void)
{
	int	i;
	int	chksum;

	if(Pass == 1)
		return;

	if(E_total==0){
		E_pc = Pc;

		return;

		}

	chksum =  E_total+3;	/* total bytes in this record */

	chksum += lobyte(E_pc);

	chksum += E_pc>>8;

	fprintf(Objfil,"S1");   /* record header preamble */

	hexout(E_total+3);	/* byte count +3 */

	hexout(E_pc>>8);	/* high byte of PC */

	hexout(lobyte(E_pc));	/* low byte of PC */

	for(i=0;i<E_total;i++){

		chksum += lobyte(E_bytes[i]);

		hexout(lobyte(E_bytes[i]));	/* data byte */

		}

	chksum =~ chksum;	/* one's complement */

	hexout(lobyte(chksum)); /* checksum */

	fprintf(Objfil,"\n");

	E_pc = Pc;

	E_total = 0;

}




/*
 *	emit --- emit a byte to code file
 */
int emit(int byte)
{

#ifdef DEBUG

	printf("%2x @ %4x\n",byte,Pc);

#endif

	if(Pass==1){

		Pc++;

		return(YES);

		}

	if(P_total < P_LIMIT)

		P_bytes[P_total++] = byte;

	E_bytes[E_total++] = byte;

	Pc++;

	if(E_total == E_LIMIT)

		f_record();

	return 0;
}

/*
 *	eword --- emit a word to code file
 */
void eword(int wd)
{
	emit(hibyte(wd));
	emit(lobyte(wd));
}



/*
 *      do_op --- process mnemonic
 *
 *	Called with the base opcode and it's class. Optr points to
 *	the beginning of the operand field.
 */
void do_op(int opcode,int classs)
{

	int     dist;   /* relative branch distance */

	int     amode;  /* indicated addressing mode */

	char	*peek;



	/* guess at addressing mode */

	peek = Optr;

	amode = OTHER;

	while( !delim(*peek) && *peek != EOS)  /* check for comma in operand field */

		if( *peek++ == ',' ){

			if( mapdn(*peek) == 'y' )

				amode = INDY;

			else

				amode = INDX;

			break;

			}

	if( *Optr == '#' ) amode = IMMED;



	yflag = 0;

	switch(classs){

		case P2INH:

			emit(PAGE2);

		case INH:                       /* inherent addressing */

			emit(opcode);

			return;

		case REL:                       /* relative branches */

			eval();

			dist = Result - (Pc+2);

			emit(opcode);

			if( (dist >127 || dist <-128) && Pass==2){

				error("Branch out of Range");

				emit(lobyte(-2));

				return;

				}

			emit(lobyte(dist));

			return;

		case LONGIMM:

			if( amode == IMMED )

				amode = LIMMED;

		case NOIMM:

			if( amode == IMMED ){

				error("Immediate Addressing Illegal");

				return;

				}

		case GEN:                       /* general addressing */

			do_gen(opcode,amode,PAGE1,PAGE1,PAGE2);

			return;

		case GRP2:

			if( amode == INDY ){

				Cycles++;

				emit(PAGE2);

				amode = INDX;

				}

			if( amode == INDX )

				do_indexed(opcode);

			else{   /* extended addressing */

				eval();

				emit(opcode+0x10);

				eword(Result);

				}

			return;

		case CPD:               /* cmpd */

			if( amode == IMMED )

				amode = LIMMED;

			if( amode == INDY )

				yflag=1;

			do_gen(opcode,amode,PAGE3,PAGE3,PAGE4);

			return;

		case XNOIMM:            /* stx */

			if( amode == IMMED ){

				error("Immediate Addressing Illegal");

				return;

				}

		case XLIMM:             /* cpx, ldx */

			if( amode == IMMED )

				amode = LIMMED;

			do_gen(opcode,amode,PAGE1,PAGE1,PAGE4);

			return;

		case YNOIMM:            /* sty */

			if( amode == IMMED ){

				error("Immediate Addressing Illegal");

				return;

				}

		case YLIMM:             /* cpy, ldy */

			if(amode == INDY)

				yflag=1;

			if( amode == IMMED )

				amode = LIMMED;

			do_gen(opcode,amode,PAGE2,PAGE3,PAGE2);

			return;

		case BTB:               /* bset, bclr */

		case SETCLR:            /* brset, brclr */

			opcode = bitop(opcode,amode,classs);

			if (amode == INDX)

				Cycles++;

			if( amode == INDY ){

				Cycles+=2;

				emit(PAGE2);

				amode = INDX;

				}

			emit(opcode);

			eval();

			emit(lobyte(Result));   /* address */

			if( amode == INDX )

				Optr += 2;      /* skip ,x or ,y */

			Optr = skip_white(Optr);

			eval();

			emit(lobyte(Result));   /* mask */

			if( classs == SETCLR )

				return;

			Optr = skip_white(Optr);

			eval();

			dist = Result - (Pc+1);

			if( (dist >127 || dist <-128) && Pass==2){

				error("Branch out of Range");

				dist = Old_pc - (Pc+1);

				}

			emit(lobyte(dist));

			return;

		default:

			fatal("Error in Mnemonic table");

		}

}



/*

 *      bitop --- adjust opcode on bit manipulation instructions

 */

int bitop(int op,int mode,int classs)
{

	if( mode == INDX || mode == INDY )

		return(op);

	if( classs == SETCLR )

		return(op-8);

	else if(classs==BTB)

		return(op-12);

	else

		fatal("bitop");

	return 0;
}



/*
 *      do_gen --- process general addressing modes
 */

void do_gen(int op,int mode,int pnorm,int px,int py)
{

	switch(mode){

	case LIMMED:

		Optr++;

		epage(pnorm);

		emit(op);

		eval();

		eword(Result);

		break;

	case IMMED:

		Optr++;

		epage(pnorm);

		emit(op);

		eval();

		emit(lobyte(Result));

		break;

	case INDY:

		if(yflag)

			Cycles += 2;

		else

			Cycles += 3;

		epage(py);

		do_indexed(op+0x20);

		break;

	case INDX:

		Cycles+=2;

		epage(px);

		do_indexed(op+0x20);

		break;

	case OTHER:

		eval();

		epage(pnorm);

		if(Force_word){

			emit(op+0x30);

			eword(Result);

			Cycles+=2;

			break;

			}

		if(Force_byte){

			emit(op+0x10);

			emit(lobyte(Result));

			Cycles++;

			break;

			}

		if(Result>=0 && Result <=0xFF){

			emit(op+0x10);

			emit(lobyte(Result));

			Cycles++;

			break;

			}

		else {

			emit(op+0x30);

			eword(Result);

			Cycles+=2;

			break;

			}

		break;

	default:

		error("Unknown Addressing Mode");

	}

}



/*
 *      do_indexed --- handle all wierd stuff for indexed addressing
 */

void do_indexed(int op)
{

	char c;



	emit(op);

	eval();

	if( *Optr++ != ',' )

		error("Syntax");

	c = mapdn(*Optr++);

	if( c != 'x' && c != 'y')

		warn("Indexed Addressing Assumed");

	if( Result < 0 || Result > 255)

		warn("Value Truncated");

	emit(lobyte(Result));

}



/*
 *      epage --- emit page prebyte
 */

void epage(int p)
{
	if( p != PAGE1 )        /* PAGE1 means no prebyte */
		emit(p);
}

/*
 *      pseudo --- pseudo op processing
 */

#define RMB     0       /* Reserve Memory Bytes         */
#define FCB     1       /* Form Constant Bytes          */
#define FDB     2       /* Form Double Bytes (words)    */
#define FCC     3       /* Form Constant Characters     */
#define ORG     4       /* Origin                       */
#define EQU     5       /* Equate                       */
#define ZMB     6       /* Zero memory bytes            */
#define FILL    7       /* block fill constant bytes    */
#define OPT     8       /* assembler option             */
#define NULL_OP 9       /* null pseudo op               */
#define PAGE    10      /* new page                     */
#define SET     11      /* Set ("re-Equate")            */

struct oper pseudo[] = {
"bsz",  PSEUDO, ZMB,    0,
"end",  PSEUDO, NULL_OP,0,
"equ",  PSEUDO, EQU,    0,
"fcb",  PSEUDO, FCB,    0,
"fcc",  PSEUDO, FCC,    0,
"fdb",  PSEUDO, FDB,    0,
"fill", PSEUDO, FILL,   0,
"nam",  PSEUDO, NULL_OP,0,
"name", PSEUDO, NULL_OP,0,
"opt",  PSEUDO, OPT,    0,
"org",  PSEUDO, ORG,    0,
"pag",  PSEUDO, PAGE,   0,
"page", PSEUDO, PAGE,   0,
"rmb",  PSEUDO, RMB,    0,
"set",  PSEUDO, SET,    0,
"spc",  PSEUDO, NULL_OP,0,
"ttl",  PSEUDO, NULL_OP,0,
"zmb",  PSEUDO, ZMB,    0
};

/*
 *      do_pseudo --- do pseudo op processing
 */

void do_pseudo(int op) /* which op */
{
        char    fccdelim;
        int     fill;        

		if( op != EQU && op != SET && *Label )
			install(Label,Pc,FALSE);

        P_force++;

        switch(op){

                case RMB:                       /* reserve memory bytes */

                        if( eval() ){

                                Pc +=  Result;

                                f_record();     /* flush out bytes */

                                }

                        else

                                error("Undefined Operand during Pass One");

                        break;

                case ZMB:                       /* zero memory bytes */

                        if( eval() )

                                while( Result-- )

                                        emit(0);

                        else

                                error("Undefined Operand during Pass One");

                        break;

                case FILL:                      /* fill memory with constant */

                        eval();

                        fill = Result;

                        if( *Optr++ != ',' )
                                error("Bad fill");
                        else{
                                Optr = skip_white(Optr);
                                eval();
                                while( Result-- )
                                        emit(fill);
                                }

                        break;

                case FCB:                       /* form constant byte(s) */

                        do{

                                Optr = skip_white(Optr);

                                eval();

                                if( Result > 0xFF ){

                                        if(!Force_byte)

                                                warn("Value truncated");

                                        Result = lobyte(Result);

                                        }

                                emit(Result);

                        }while( *Optr++ == ',' );

                        break;

                case FDB:                       /* form double byte(s) */

                        do{

                                Optr = skip_white(Optr);

                                eval();

                                eword(Result);

                        }while( *Optr++ == ',' );

                        break;

                case FCC:                       /* form constant characters */

                        if(*Operand==EOS)

                                break;

                        fccdelim = *Optr++;

                        while( *Optr != EOS && *Optr != fccdelim)

                                emit(*Optr++);

                        if(*Optr == fccdelim)

                                Optr++;

                        else

                                error("Missing Delimiter");

                        break;

                case ORG:                       /* origin */

                        if( eval() ){

                                Old_pc = Pc = Result;

                                f_record();     /* flush out any bytes */

                                }

                        else

                                error("Undefined Operand during Pass One");

                        break;

                case EQU:                       /* equate */

                        if(*Label==EOS){

                                error("EQU requires label");

                                break;

                                }

                        if( eval() ){

				install(Label,Result,FALSE);

                                Old_pc = Result;        /* override normal */

                                }

                        else

                                error("Undefined Operand during Pass One");

                        break;

		case SET:                       /* set */

                        if(*Label==EOS){

				error("SET requires label");

                                break;

                                }

                        if( eval() ){

				install(Label,Result,TRUE);

                                Old_pc = Result;        /* override normal */

                                }

                        else

                                error("Undefined Operand during Pass One");

                        break;

                case OPT:                       /* assembler option */

                        P_force=0;

                        if( head(Operand,"l") )

                                Lflag=1;

                        else if( head(Operand,"nol"))

                                Lflag=0;

                        else if( head(Operand,"c")){

                                Cflag=1;

                                Ctotal=0;

                                }

                        else if( head(Operand,"noc"))

                                Cflag=0;

                        else if( head(Operand,"contc")){

                                Cflag=1;

                                }

                        else if ( head(Operand,"s"))

                                Sflag = 1;

                        else if ( head(Operand,"cre"))

                                CREflag = 1;

                        else

                                error("Unrecognized OPT");

                        break;

                case PAGE:                      /* go to a new page */

                        P_force=0;

                        N_page = 1;

                        if (Pass == 2 )

                         if (Lflag)  

                          {

                           fprintf (lstfile, "\f");

                           fprintf (lstfile, "%-10s",Argv[Cfn]);

                           fprintf (lstfile, "                                   ");

                           fprintf (lstfile, "page %3d\n",Page_num++);

                          }

                        break;

                case NULL_OP:                   /* ignored psuedo ops */

                        P_force=0;

                        break;

                default:

                        fatal("Pseudo error");

                }

}

/*
 *      eval --- evaluate expression
 *
 *      an expression is constructed like this:
 *
 *      expr ::=  expr + term |
 *                expr - term ;
 *                expr * term ;
 *                expr / term ;
 *                expr | term ;
 *                expr & term ;
 *                expr % term ;
 *                expr ^ term ;
 *                expr < term ;
 *                expr > term ;
 *
 *      term ::=  symbol |
 *                * |
 *                constant ;
 *
 *      symbol ::=  string of alphanumerics with non-initial digit
 *
 *      constant ::= hex constant |
 *                   binary constant |
 *                   octal constant |
 *                   decimal constant |
 *                   ascii constant;
 *
 *      hex constant ::= '$' {hex digits};
 *
 *      octal constant ::= '@' {octal digits};
 *
 *      binary constant ::= '%' { 1 | 0 };
 *
 *      decimal constant ::= {decimal digits};
 *
 *      ascii constant ::= ''' any printing char;
 *
 */

int eval(void)
{

	int     left,right;     /* left and right terms for expression */
	char    o;              /* operator character */



#ifdef DEBUG

	printf("Evaluating %s\n",Optr);

#endif

	Force_byte = NO;

	Force_word = NO;

	if(*Optr=='<'){

		Force_byte++;

		Optr++;

		}

	else if(*Optr=='>'){

		Force_word++;

		Optr++;

		}

	left = get_term();         /* pickup first part of expression */



	while( is_op(*Optr)){

		o = *Optr++; /* pickup connector and skip */

		right = get_term();     /* pickup current rightmost side */

		switch(o){

			case '+': left += right; break;

			case '-': left -= right; break;

			case '*': left *= right; break;

			case '/': left /= right; break;

			case '|': left |= right; break;

			case '&': left &= right; break;

			case '%': left %= right; break;

			case '^': left = left^right; break;

			case '>': left >>= right; break;

			case '<': left <<= right; break;

			}

		}



	Result= left;

#ifdef DEBUG

	printf("Result=%x\n",Result);

	printf("Force_byte=%d  Force_word=%d\n",Force_byte,Force_word);

#endif

	return(YES);

}


/*
 *      is_op --- is character an expression operator?
 */

int is_op(char c)
{

	if( any(c,"+-*/&%|^<>"))

		return(YES);

	return(NO);

}





/*
 *      get_term --- evaluate a single item in an expression
 */

int get_term(void)
{
	char    hold[MAXBUF];
	char    *tmp;
	int     val = 0;        /* local value being built */
	int     minus;          /* unary minus flag */

	if( *Optr == '-' ){

		Optr++;

		minus =YES;

		}

	else

		minus = NO;



	while( *Optr == '#' ) Optr++;



	/* look at rest of expression */



	if(*Optr=='%'){ /* binary constant */

		Optr++;

		while( any(*Optr,"01"))

			val = (val * 2) + ( (*Optr++)-'0');

		}

	else if(*Optr=='@'){ /* octal constant */

		Optr++;

		while( any(*Optr,"01234567"))

			val = (val * 8) + ((*Optr++)-'0');

		}

	else if(*Optr=='$'){ /* hex constant */

		Optr++;

		while( any(*Optr,"0123456789abcdefABCDEF"))

			if( *Optr > '9' )

				val = (val * 16) + 10 + (mapdn(*Optr++)-'a');

			else

				val = (val * 16) + ((*Optr++)-'0');

		}

	else if( any(*Optr,"0123456789")){ /* decimal constant */

		while(*Optr >= '0' && *Optr <= '9')

			val = (val * 10) + ( (*Optr++)-'0');

		}

	else if(*Optr=='*'){    /* current location counter */

		Optr++;

		val = Old_pc;

		}

	else if(*Optr=='\''){   /* character literal */

		Optr++;

		if(*Optr == EOS)

			val = 0;

		else

			val = *Optr++;

		}

	else if( alpha(*Optr) ){ /* a symbol */

		tmp = hold;     /* collect symbol name */

		while(alphan(*Optr))

			*tmp++ = *Optr++;

		*tmp = EOS;

		if(lookup(hold) != NULL)

			val = Last_sym;

		else{

			if(Pass==1){    /* forward ref here */

				fwdmark();

				if( !Force_byte )

					Force_word++;

				val = 0;

				}

			}

		if(Pass==2 && Line_num==F_ref && Cfn==Ffn){

			if( !Force_byte  )

				Force_word++;

			fwdnext();

			}

		}

	else

		/* none of the above */

		val = 0;



	if(minus)

		return(-val);

	else

		return(val);

}

/*
 *      install --- add a symbol to the table
 */

int install(char *str, int val, int override)
{
    struct link *lp;
	struct nlist *np,*p,*backp;
	int     i;

	if( !alpha(*str) ){

		error("Illegal Symbol Name");

		return(NO);

		}

	if( (np = lookup(str)) != NULL ){

	    if (override) {

		np -> def = val;

		return (YES);

		}

	    else {

		if( Pass==2 ){

			if( np->def == val )

				return(YES);

			else{

				error("Phasing Error");

				return(NO);

				}

			}

		error("Symbol Redefined");

		return(NO);

		}

	    }

	/* enter new symbol */

#ifdef DEBUG

	printf("Installing %s as %d\n",str,val);

#endif

	np = (struct nlist *) alloc(sizeof(struct nlist));

	if( np == (struct nlist *)ERR ){

		error("Symbol table full");

		return(NO);

		}

	np->name = alloc(strlen(str)+1);

	if( np->name == (char *)ERR ){

		error("Symbol table full");

		return(NO);

		}

	strcpy(np->name,str);

	np->def = val;

	np->Lnext = NULL;  

        np->Rnext = NULL;

           lp = (struct link *) alloc(sizeof(struct link));

           np->L_list = lp;

           lp->L_num = Line_num;

           lp->next = NULL;

        p = root;

          backp = NULL;

           while (p != NULL) 

            {

              backp = p;

              i = strcmp (str,p->name);

               if (i<0)

                   p=p->Lnext;

                  else p=p->Rnext;

            }

          if (backp == NULL)

              root = np;

             else if (strcmp(str,backp->name)<0)

                  backp->Lnext = np;

                 else backp->Rnext = np;

          return (YES);  

}



/*
 *      lookup --- find string in symbol table
 */

struct nlist *lookup(char *name)
{

	struct nlist *np;

	int     i;



        np = root;

         while (np != NULL)

          {

            i = strcmp(name,np->name);

             if (i == 0)

               {

                 Last_sym = np->def;

                 return (np);

               }

             else if (i < 0)

                 np = np->Lnext;

                else np = np->Rnext;

          }

      Last_sym = 0;

      if (Pass == 2)

          error ("symbol Undefined on pass 2");

        return (NULL); 

}

#define NMNE (sizeof(table)/ sizeof(struct oper))
#define NPSE (sizeof(pseudo)/ sizeof(struct oper))

/*
 *      mne_look --- mnemonic lookup
 *
 *      Return pointer to an oper structure if found.
 *      Searches both the machine mnemonic table and the pseudo table.
 */

struct oper *mne_look(char *str)
{

	struct oper *low,*high,*mid;

	int     cond;



	/* Search machine mnemonics first */

	low =  &table[0];

	high = &table[ NMNE-1 ];

	while (low <= high){

		mid = low + (high-low)/2;

		if( ( cond = strcmp(str,mid->mnemonic)) < 0)

			high = mid - 1;

		else if (cond > 0)

			low = mid + 1;

		else

			return(mid);

	}



	/* Check for pseudo ops */

	low =  &pseudo[0];

	high = &pseudo[ NPSE-1 ];

	while (low <= high){

		mid = low + (high-low)/2;

		if( ( cond = strcmp(str,mid->mnemonic)) < 0)

			high = mid - 1;

		else if (cond > 0)

			low = mid + 1;

		else

			return(mid);

	}



	return(NULL);

}

/*
 *	fatal --- fatal error handler
 */
void  fatal(char	*str)
{
	fprintf(lstfile, "%s\n",str);	
	fatalerr=1;
}



/*
 *	error --- error in a line
 *			print line number and error
 */
void error(char	*str)
{

	if(N_files > 1)

		fprintf(lstfile, "%s,",Argv[Cfn]);

	fprintf(lstfile, "%d| ",Line_num);

	fprintf(lstfile, "%s\n",str);

	Err_count++;

}

/*
 *	warn --- trivial error in a line
 *			print line number and error
 */
void warn(char	*str)
{

	if(N_files > 1)

		fprintf(lstfile, "%s,",Argv[Cfn]);

	fprintf(lstfile, "%d: ",Line_num);

	fprintf(lstfile, "Warning --- %s\n",str);

}

#define	FILEMODE	0644	/* file creat mode */
#define	UPDATE		2	/* file open mode */
#define	ABS		0	/* absolute seek */

int	Forward =0;		/* temp file's file descriptor	*/
char	Fwd_name[] = { "Fwd_refs" } ;

/*
 *      fwdinit --- initialize forward ref file
 */
void fwdinit(void)
{

	Forward = creat(Fwd_name,FILEMODE);

	if(Forward <0)

		fatal("Can't create temp file");

	close(Forward); /* close and reopen for reads and writes */

	Forward = open(Fwd_name,UPDATE);

	if(Forward <0)

		fatal("Forward ref file has gone.");

#ifndef DEBUG

	unlink(Fwd_name);

#endif

}



/*
 *      fwdreinit --- reinitialize forward ref file
 */
void fwdreinit(void)
{
	F_ref   = 0;

	Ffn     = 0;

	lseek(Forward,0L,ABS);   /* rewind forward refs */

	read(Forward,&Ffn,sizeof(Ffn));

	read(Forward,&F_ref,sizeof(F_ref)); /* read first forward ref into mem */

#ifdef DEBUG

	printf("First fwd ref: %d,%d\n",Ffn,F_ref);

#endif

}



/*

 *      fwdmark --- mark current file/line as containing a forward ref

 */

void fwdmark(void)
{
	write(Forward,&Cfn,sizeof(Cfn));
	write(Forward,&Line_num,sizeof(Line_num));
}



/*
 *      fwdnext --- get next forward ref
 */
void fwdnext(void)
{

	int stat;



	stat = read(Forward,&Ffn,sizeof(Ffn));

#ifdef DEBUG

	printf("Ffn stat=%d ",stat);

#endif

	stat = read(Forward,&F_ref,sizeof(F_ref));

#ifdef DEBUG

	printf("F_ref stat=%d  ",stat);

#endif

	if( stat < 2 ){

		F_ref=0;Ffn=0;

		}

#ifdef DEBUG

	printf("Next Fwd ref: %d,%d\n",Ffn,F_ref);

#endif

}

/*
 *	print_line --- pretty print input line
 */

void print_line(void)
{

	int	i;

	register char *ptr;



	fprintf (lstfile, "%04d ",Line_num);

	if(P_total || P_force)
		fprintf(lstfile, "%04x",Old_pc);
	else
		fprintf(lstfile, "    ");

	for(i=0;i<P_total && i<6;i++)
		fprintf(lstfile, " %02x",lobyte(P_bytes[i]));

	for(;i<6;i++)
		fprintf(lstfile, "   ");

	fprintf(lstfile, "  ");


	if(Cflag){
		if(Cycles)
			fprintf(lstfile, "[%2d ] ",Cycles);
		else
			fprintf(lstfile, "      ");
		}

	ptr = Line;

	while( *ptr != '\n' )   /* just echo the line back out */
		fprintf(lstfile,"%c", *ptr++);

	for(;i<P_total;i++){

		if( i%6 == 0 )

			fprintf(lstfile, "\n    ");

		fprintf(lstfile, " %02x",lobyte(P_bytes[i]));
		}
	fprintf(lstfile, "\n");
}

/*
 *  stable --- prints the symbol table in alphabetical order
 */
void stable(struct nlist *ptr)
{

  if (ptr != NULL)

    {

      stable (ptr->Lnext);

        printf ("%-10s %04x\n",ptr->name,ptr->def);

      stable (ptr->Rnext);

    }

}


/*

 *  cross  --  prints the cross reference table 

 */
void cross(struct nlist *point)
{

struct link *tp;

int i = 1;

  if (point != NULL)

    {

      cross (point->Lnext);

        printf ("%-10s %04x *",point->name,point->def);

         tp = point->L_list;

          while (tp != NULL)

           {

             if (i++>10)

              {

               i=1;

               printf("\n                      ");

              }

              printf ("%04d ",tp->L_num);

               tp = tp->next;

           }

         printf ("\n");

      cross (point->Rnext);

    }

}
void initialize(void)
{
        int     i = 0;

#ifdef DEBUG
        printf("Initializing\n");
#endif
        Err_count = 0;
        Pc        = 0;
        Pass      = 1;
        Lflag     = 1;
        Cflag     = 0;
        Ctotal    = 0;
        Sflag     = 0;
        CREflag   = 0;
        N_page    = 0;
        Line[MAXBUF-1] = NEWLINE;

        fwdinit();      /* forward ref init */
}

void re_init(void)
{
#ifdef DEBUG
        printf("Reinitializing\n");
#endif
        Pc      = 0;
        E_total = 0;
        P_total = 0;
        Ctotal  = 0;
        N_page  = 0;
        fwdreinit();
}


void make_pass(void)
{
#ifdef DEBUG
        printf("Pass %d\n",Pass);
#endif
        while( fgets(Line,MAXBUF-1,Fd) != (char *)NULL ){
                Line_num++;
                P_force = 0;    /* No force unless bytes emitted */
                N_page = 0;
                if(parse_line())
                        process();
                if(Pass == 2 && Lflag && !N_page)
                        print_line();
                P_total = 0;    /* reset byte count */
                Cycles = 0;     /* and per instruction cycle count */
				if(fatalerr)break;
                }
        f_record();
}


/*
 *      parse_line --- split input line into label, op and operand
 */
int parse_line(void)
{
        register char *ptrfrm = Line;
        register char *ptrto = Label;

        if( *ptrfrm == '*' || *ptrfrm == ';' || *ptrfrm == '\n' )
                return(0);      /* a comment line */

        while( (delim(*ptrfrm)== NO) && (*ptrfrm!=';'))
                *ptrto++ = *ptrfrm++;
        if(*--ptrto != ':')ptrto++;     /* allow trailing : */
        *ptrto = EOS;

        ptrfrm = skip_white(ptrfrm);

        ptrto = Op;
        while( (delim(*ptrfrm) == NO) && (*ptrfrm!=';'))
                *ptrto++ = mapdn(*ptrfrm++);
        *ptrto = EOS;

        ptrfrm = skip_white(ptrfrm);

        ptrto = Operand;
        while( (*ptrfrm != NEWLINE) && (*ptrfrm!=';'))
                *ptrto++ = *ptrfrm++;
        *ptrto = EOS;

#ifdef DEBUG
        printf("Label-%s-\n",Label);
        printf("Op----%s-\n",Op);
        printf("Operand-%s-\n",Operand);
#endif
        return(1);
}

/*
 *      process --- determine mnemonic class and act on it
 */
void process(void)
{
        register struct oper *i;

        Old_pc = Pc;            /* setup `old' program counter */
        Optr = Operand;         /* point to beginning of operand field */

        if(*Op==EOS){           /* no mnemonic */
                if(*Label != EOS)
                        install(Label,Pc, FALSE);
                }
        else if( (i = mne_look(Op))== NULL)
                error("Unrecognized Mnemonic");
        else if( i->classs == PSEUDO )
                do_pseudo(i->opcode);
        else{
                if( *Label )install(Label,Pc, FALSE);
                if(Cflag)Cycles = i->cycles;
                do_op(i->opcode,i->classs);
                if(Cflag)Ctotal += Cycles;
                }
}

/*

 *      MC68HC11 specific processing

 */


